/****************************************************************************
 Jed Plugin: CreateVertex

 This plugin creates a new vertex on the current sector
 ****************************************************************************/

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <stdio.h>
#include "jed_plugins.h"

/****************************
	My function prototypes:
 ****************************/

void getSectorMidpoint(int sec, double *x, double *y, double *z);
void getSurfaceMidpoint(int sec, int surf, double *x, double *y, double *z);
void getMidpoint(double *x_arr, double *y_arr, double *z_arr, int size, double *x, double *y, double *z);

/****************************/

HINSTANCE hInst = 0;

IJED *jed = 0;
IJEDLevel *level = 0;


bool APIENTRY DllMain( HANDLE hModule, 
					   DWORD  ul_reason_for_call, 
					   LPVOID lpReserved
					 )
{
	hInst = (HINSTANCE)hModule;
	switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return true;
}

// this is what jed calls when you activate the plugin
int _stdcall JEDPluginLoadStdCall(IJED *j)
{
	bool debug = false;
	jed = j;
	level = jed->GetLevel();
	int cursc;												// sector for which the new vertex is for
	double newx, newy, newz;								// coords for new vertex

	cursc = jed->GetCurSC();								// get current sector

	if(jed->GetMapMode() == MM_VX)							// vertex mode
	{
		int numvx;
		numvx = jed->GetNMultiselected(MM_VX);				// get number of multiselected vertices
		if(numvx < 1)										// if we have no multiselected vertices
		{
			if(debug) MessageBox(0, "No multiselected vertices.", 0, MB_OK);
			getSectorMidpoint(cursc, &newx, &newy, &newz);	// center of current sector
		}
		else if(numvx == 1)									// exactly one multiselected vertex...
		{
			int s0, s1, v0, v1;
			if(debug) MessageBox(0, "Exactly 1 multiselected vertex.", 0, MB_OK);
			double x[2], y[2], z[2];
			jed->GetCurVX(&s0, &v0);
			jed->GetSelectedVX(0, &s1, &v1);
			level->SectorGetVertex(s0, v0, &x[0], &y[0], &z[0]);
			level->SectorGetVertex(s1, v1, &x[1], &y[1], &z[1]);
			getMidpoint(x, y, z, 2, &newx, &newy, &newz);
		}
		else if(numvx > 1)									// 2 or more multiselected vertices...
		{
			if(debug) MessageBox(0, "Multiple multiselected vertices.", 0, MB_OK);
			double *x, *y, *z;
			x = new double[numvx];
			y = new double[numvx];
			z = new double[numvx];
			if(x == NULL || y == NULL || z == NULL)
			{
				delete x;
				delete y;
				delete z;
				newx = 0;
				newy = 0;
				newz = 0;
			}
			else
			{
				int sc, vx;
				for(int i = 0; i < numvx; i++)
				{
					jed->GetSelectedVX(i, &sc, &vx);
					level->SectorGetVertex(sc, vx, &x[i], &y[i], &z[i]);
				}
				getMidpoint(x, y, z, numvx, &newx, &newy, &newz);
				delete x;
				delete y;
				delete z;
			}
		}
		else												// dunno how we would get here, but add it for failsafe
		{
			MessageBox(0, "An unexpected error has just occurred.", 0, MB_OK);
			return 0;
		}
	}
	else if(jed->GetMapMode() == MM_SC)						// Sector Mode
	{
		if(debug) MessageBox(0, "Sector mode -- create vertex at center of sector.", 0, MB_OK);
		getSectorMidpoint(cursc, &newx, &newy, &newz);		// center of current sector
		jed->SetMapMode(MM_VX);								// set to vertex mode
	}
	else if(jed->GetMapMode() == MM_SF)						// Surface Mode
	{
		if(debug) MessageBox(0, "Surface mode -- create vertex at center of surface.", 0, MB_OK);
		int cursf, sc;
		cursf = jed->GetCurSF(&sc, &cursf);
		getSurfaceMidpoint(cursc, cursf, &newx, &newy, &newz);	// center of current surface
		jed->SetMapMode(MM_VX);								// set to vertex mode
	}
	else
	{
		MessageBox(0, "Must be in either Sector, Surface, or Vertex mode to use plugin.", 0, MB_OK);
		return 1;
	}

	if(debug) MessageBox(0, "Creating vertex...", 0, MB_OK);

	jed->StartUndo("CreateVertex");						// save undo
	jed->SaveUndoForSector(cursc, uc_changed, 0);		// save undo state for sector

	int newvx;											// index for the new vertex
	// insert the vertex at the calculated coords
	newvx = level->SectorAddVertex(cursc, newx, newy, newz);	// add the vertex
	jed->SetCurVX(cursc, newvx);						// make the new vertex be the current vertex
	level->SectorUpdate(cursc);							// update the sector
	jed->UpdateMap();									// update map
	return 1;
}


void getMidpoint(double *x_arr, double *y_arr, double *z_arr, int size, double *x, double *y, double *z)
{
	double xmin, ymin, zmin, xmax, ymax, zmax;
	if(size <= 0)
	{
		*x = 0;
		*y = 0;
		*z = 0;
		return;
	}
	xmin = x_arr[0];
	ymin = y_arr[0];
	zmin = z_arr[0];
	xmax = xmin;
	ymax = ymin;
	zmax = zmin;
	for(int i = 1; i < size; i++)
	{
		if(x_arr[i] < xmin)
			xmin = x_arr[i];
		if(x_arr[i] > xmax)
			xmax = x_arr[i];

		if(y_arr[i] < ymin)
			ymin = y_arr[i];
		if(y_arr[i] > ymax)
			ymax = y_arr[i];

		if(z_arr[i] < zmin)
			zmin = z_arr[i];
		if(z_arr[i] > zmax)
			zmax = z_arr[i];
	}
	*x = (xmin + xmax) / 2;
	*y = (ymin + ymax) / 2;
	*z = (zmin + zmax) / 2;
	return;
}

void getSectorMidpoint(int sec, double *x, double *y, double *z)
{
	int v;
	v = level->SectorNVertices(sec);
	if(v < 1)										// if sector has no vertices
	{
		*x = 0;										// default to the origin
		*y = 0;
		*z = 0;
		return;
	}

	double *x2, *y2, *z2;
	x2 = new double[v];
	y2 = new double[v];
	z2 = new double[v];
	if(x2 == NULL || y2 == NULL || z2 == NULL)		// could not allocate memory
	{
		delete x2;									// deallocate
		delete y2;
		delete z2;
		*x = 0;										// return the origin
		*y = 0;
		*z = 0;
		return;
	}
	for(int i = 0; i < v; i++)
	{
		level->SectorGetVertex(sec, i, &x2[i], &y2[i], &z2[i]);
	}
	getMidpoint(x2, y2, z2, v, x, y, z);
	delete x2;										// deallocate
	delete y2;
	delete z2;
	return;
}
void getSurfaceMidpoint(int sec, int surf, double *x, double *y, double *z)
{
	if(surf >= level->SectorNSurfaces(sec))			// if surface index >= number of surfaces
	{
		*x = 0;										// don't know what surface is wanted, so just return the origin
		*y = 0;
		*z = 0;
		return;
	}

	int v;
	v = level->SurfaceNVertices(sec, surf);
	if(v < 1)										// if surface has no vertices
	{
		*x = 0;										// default to the origin
		*y = 0;
		*z = 0;
		return;
	}

	int vx;

	double *x2, *y2, *z2;
	x2 = new double[v];
	y2 = new double[v];
	z2 = new double[v];
	if(x2 == NULL || y2 == NULL || z2 == NULL)		// could not allocate memory
	{
		delete x2;									// deallocate
		delete y2;
		delete z2;
		*x = 0;										// return the origin
		*y = 0;
		*z = 0;
		return;
	}

	// find all vertices on surface and compile into an array of coords
	for(int i = 0; i < v; i++)
	{
		vx = level->SurfaceGetVertexNum(sec, surf, i);
		level->SectorGetVertex(sec, vx, &x2[i], &y2[i], &z2[i]);
	}
	
	// calculate the center of the surface
	getMidpoint(x2, y2, z2, v, x, y, z);

	return;
}

